/*
 * StringProtector – string handling tool for JETERS components 
 * Copyright (C) 2007  Tobias Knerr
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA
 * 
 */

import java.util.regex.Pattern;
import java.util.Hashtable;
import java.util.Iterator;


/**
 * tool that will "protect" parts of a String
 * identified by regular expressions by replacing
 * them with an ID (positive integer) and some
 * strings specified in the constructor.
 * 
 * The number of protections is limited to the
 * number of possible non-negative values in a Long.
 */
class StringProtector {

	/** begin of a protected part of the string */
	private String startMarker;

	/** end of a protected part of the string */
	private String endMarker;	
	
	/** ID that will be used for the next protection */
	private Long nextID = 0l;
	
	/** 
	 * hashtable containing the replaced String
	 * for each ID.
	 */
	private Hashtable<Long,String> protHash
		= new Hashtable<Long,String>();
	
	/**
	 * creates a new string protector
	 * 
	 * @param startMarker  String that will mark the
	 *                     beginning of a protected part
	 * @param endMarker    String that will mark the
	 *                     end of a protected part
	 */
	public StringProtector(String startMarker,
			               String endMarker) {
		
		this.startMarker = startMarker;
		this.endMarker   = endMarker;
		
	}
	
	/**
	 * replaces all occurrences of the regular
	 * expression given as vararg parameter
	 * 'expressions' in the String s
	 * 
	 * @param s            string to create a 
	 *                     protected version from
	 * @param expressions  regular expr.s to protect
	 * @return             s with protections applied
	 */	 
	public String protect(String s,
			               String... expressions) {
					
		/*TODO:  erlauben, Teile der
		 *  Bereiche nicht zu schützen
		 *  (z.B. in Weblinks den Namen
		 *  nach der Adresse) ++*/
				
		String result = s;						
				
		for (String exp : expressions) {
			
			int flags = Pattern.MULTILINE
			            | Pattern.DOTALL;			
			
			Pattern pattern
				= Pattern.compile("("+exp+")",
						          flags);
			
			/* replace all occurrences of the exp. */
			
			while( pattern.matcher(result).find() ){
			
				Long id = getUnusedID();
				
				result = pattern.matcher(result)
					     .replaceFirst(startMarker + id
					    		       + "$1" + id + endMarker);			
				
				//get protected substring
				
				int firstIndexOfProtectedText 
					= result.indexOf(startMarker+id)
					  + ((startMarker+id).length());
				int firstIndexAfterProtectedText 
					= result.indexOf(id + endMarker,
							         firstIndexOfProtectedText);
				
				String protectedText
					= result.substring(firstIndexOfProtectedText,
							           firstIndexAfterProtectedText);
								
				result = Pattern.compile(startMarker+id+".*?"+id+endMarker,
						                 flags).matcher(result)
						 .replaceFirst(startMarker+id+endMarker);

				//add protected substring to hashtable
				
				protHash.put(id,protectedText);
			
			}
		
		}	
		
		return result;
		
	}

	/**
	 * replaces the protection marks
	 * with the original text;
	 * protected text will no longer be known
	 * to this StringProtector!
	 * 
	 * @param s  String to remove protection marks from
	 * @return   s with original Strings instead of prots
	 */	
	public String unprotect(String s){
				
		String result = s;
		
		boolean complete = false;
		
		/* 
		 * repeat until there are no more changes
		 * (necessary because original texts put
		 * in for protection marks might
		 * themselves contain protection marks) 
		 */
		while( !complete ){
			
			complete = true;
			
			/*
			 * TODO: more efficient algorithm:
			 * search for id markers in the text
			 * and get the string for it
			 */
			
			Iterator<Long> iterator
				= protHash.keySet().iterator();

			while( iterator.hasNext() ){
				
				Long id = iterator.next();
				
				String textBeforeReplacements = result;
											
				String protectedText = protHash.get(id);
							
				if( Pattern.compile(startMarker+id+endMarker)
						.matcher(result).find() ){				

					/* replace protection marker with original text 
					 * (note: since replace does not use regex strings, there
					 * are no problems with special characters in protectedText) 
					 */
					result = result.replace(
							startMarker+id+endMarker,
							protectedText);
					
					if (!result.equals(textBeforeReplacements) ) {

						iterator.remove();
						complete = false;

						/*TODO: remove if thoroughly tested
						 * (backup test whether the original text
						 * is in the string again)
						 */
						if ( !result.contains(protectedText) ) {
							throw new Error("StringProtector encountered a"
									+ "error while unprotecting parts of the text."
									+ "\nPlease consider sending a bug report.\n"
									+ "Error description:\n"
									+ "id: " + id + "\nprotected text:"
									+ protectedText + "\nresult:" + result);
						}

					}
				
				}
				
			}
			
		}

		return result;		
				
	}
	
	/**
	 * returns an unique, unused ID
	 * 
	 *  @return  unique id; null if all
	 *           possible ids have been used
	 */
	private Long getUnusedID(){
	
		Long id = nextID;
		
		if (id != null && id != Long.MAX_VALUE) {
			nextID = id + 1;			
		}
		
		return id; 
			 
	}
	
}
